@@ -1,21 +1,19 @@
 """This module provides methods for handling queries."""
 import re
-from typing import List, Dict, Set, Any, TypeVar, Callable, Optional
+from typing import List, Dict, Set, Any, Tuple, TypeVar, Callable, Optional
 from urllib.parse import quote
 from datetime import datetime
 
 from ga4gh.vrsatile.pydantic.vrs_models import VRSTypes
 from ga4gh.vrsatile.pydantic.vrsatile_models import GeneDescriptor, Extension
-from botocore.exceptions import ClientError
-from boto3.dynamodb.conditions import Key
 from ga4gh.vrs import models
 from ga4gh.core import ga4gh_identify
 
 from gene import logger
 from gene import NAMESPACE_LOOKUP, PREFIX_LOOKUP, ITEM_TYPES
-from gene.database import Database
-from gene.schemas import BaseGene, Gene, SourceMeta, MatchType, SourceName, \
-    ServiceMeta, SourcePriority, NormalizeService, SearchService, \
+from gene.database import AbstractDatabase, DatabaseReadException
+from gene.schemas import BaseGene, Gene, NamespacePrefix, RefType, MatchType, \
+    SourceName, ServiceMeta, SourcePriority, NormalizeService, SearchService, \
     GeneTypeFieldName, UnmergedNormalizationService, MatchesNormalized, \
     BaseNormalizationService
 from gene.version import __version__
@@ -34,19 +32,15 @@ class QueryHandler:
     and normalizes query input.
     """
 
-    def __init__(self, db_url: str = '', db_region: str = 'us-east-2'):
-        """Initialize QueryHandler instance.
-
-        :param str db_url: URL to database source.
-        :param str db_region: AWS default region.
-        """
-        self.db = Database(db_url=db_url, region_name=db_region)
+    def __init__(self, database: AbstractDatabase) -> None:
+        """Initialize QueryHandler instance."""
+        self.db = database
 
     @staticmethod
     def emit_warnings(query_str: str) -> List:
         """Emit warnings if query contains non breaking space characters.
 
-        :param str query_str: query string
+        :param query_str: query string
         :return: List of warnings
         """
         warnings = []
@@ -63,28 +57,11 @@ class QueryHandler:
             )
         return warnings
 
-    def fetch_meta(self, src_name: str) -> SourceMeta:
-        """Fetch metadata for src_name.
-
-        :param str src_name: name of source to get metadata for
-        :return: SourceMeta object containing source metadata
-        """
-        if src_name in self.db.cached_sources.keys():
-            return self.db.cached_sources[src_name]
-        else:
-            try:
-                db_response = self.db.metadata.get_item(Key={'src_name':
-                                                             src_name})
-                response = SourceMeta(**db_response['Item'])
-                self.db.cached_sources[src_name] = response
-                return response
-            except ClientError as e:
-                logger.error(e.response['Error']['Message'])
-
     @staticmethod
     def _transform_sequence_location(loc: Dict) -> models.SequenceLocation:
         """Transform a sequence location to VRS sequence location
-        :param Dict loc: Sequence location
+
+        :param loc: Sequence location
         :return: VRS sequence location
         """
         return models.SequenceLocation(
@@ -99,7 +76,7 @@ class QueryHandler:
     def _transform_chromosome_location(loc: Dict) -> models.ChromosomeLocation:
         """Transform a chromosome location to VRS chromosome location
 
-        :param Dict loc: Chromosome location
+        :param loc: Chromosome location
         :return: VRS chromosome location
         """
         transformed_loc = models.ChromosomeLocation(
@@ -114,20 +91,21 @@ class QueryHandler:
 
     def _transform_location(self, loc: Dict) -> Dict:
         """Transform a sequence/chromosome location to VRS sequence/chromosome location
-        :param Dict loc: Sequence or Chromosome location
+
+        :param loc: Sequence or Chromosome location
         :return: VRS sequence or chromosome location represented as a dictionary
         """
         if loc["type"] == VRSTypes.SEQUENCE_LOCATION:
-            loc = self._transform_sequence_location(loc)
+            transformed_loc = self._transform_sequence_location(loc)
         else:
-            loc = self._transform_chromosome_location(loc)
-        loc._id = ga4gh_identify(loc)
-        return loc.as_dict()
+            transformed_loc = self._transform_chromosome_location(loc)
+        transformed_loc._id = ga4gh_identify(transformed_loc)
+        return transformed_loc.as_dict()
 
     def _transform_locations(self, record: Dict) -> Dict:
         """Transform gene locations to VRS Chromosome/Sequence Locations
 
-        :param Dict record: original record
+        :param record: original record
         :return: record with transformed locations attributes, if applicable
         """
         record_locations = list()
@@ -137,25 +115,36 @@ class QueryHandler:
         record["locations"] = record_locations
         return record
 
+    def _get_src_name(self, concept_id: str) -> SourceName:
+        """Get source name enum from ID.
+
+        :param concept_id: candidate concept ID string to check
+        :return: SourceName option
+        :raise: ValueError if unrecognized ID provided
+        """
+        if concept_id.startswith(NamespacePrefix.ENSEMBL.value):
+            return SourceName.ENSEMBL
+        elif concept_id.startswith(NamespacePrefix.NCBI.value):
+            return SourceName.NCBI
+        elif concept_id.startswith(NamespacePrefix.HGNC.value):
+            return SourceName.HGNC
+        else:
+            raise ValueError("Invalid or unrecognized concept ID provided")
+
     def add_record(self,
                    response: Dict[str, Dict],
                    item: Dict,
-                   match_type: MatchType) -> (Dict, str):
+                   match_type: MatchType):
         """Add individual record (i.e. Item in DynamoDB) to response object
 
-        :param Dict[str, Dict] response: in-progress response object to return
-            to client
-        :param Dict item: Item retrieved from DynamoDB
-        :param MatchType match_type: match type for query
-        :return: Tuple containing updated response object, and string
-            containing name of the source of the match
+        :param response: in-progress response object to return to client
+        :param item: Item retrieved from DynamoDB
+        :param match_type: match type for query
         """
-        del item['label_and_type']
-        # DynamoDB Numbers get converted to Decimal
         item = self._transform_locations(item)
         item["match_type"] = match_type
         gene = Gene(**item)
-        src_name = item['src_name']
+        src_name = item["src_name"]
 
         matches = response['source_matches']
         if src_name not in matches.keys():
@@ -163,46 +152,45 @@ class QueryHandler:
         elif matches[src_name] is None:
             matches[src_name] = {
                 'records': [gene],
-                'source_meta_': self.fetch_meta(src_name)
+                'source_meta_': self.db.get_source_metadata(src_name)
             }
         else:
             matches[src_name]['records'].append(gene)
 
-        return response, src_name
-
     def fetch_record(self, response: Dict[str, Dict], concept_id: str,
                      match_type: MatchType) -> None:
         """Add fetched record to response
 
-        :param Dict[str, Dict] response: in-progress response object to return
-            to client.
-        :param str concept_id: Concept id to fetch record for.
-            Should be all lower-case.
-        :param MatchType match_type: match type for record
+        :param response: in-progress response object to return to client.
+        :param concept_id: Concept id to fetch record for. Should be all lower-case.
+        :param match_type: match type for record
         """
         try:
-            pk = f'{concept_id}##identity'
-            filter_exp = Key('label_and_type').eq(pk)
-            result = self.db.genes.query(KeyConditionExpression=filter_exp)
-            match = result['Items'][0]
-            self.add_record(response, match, match_type)
-        except ClientError as e:
-            logger.error(e.response['Error']['Message'])
+            match = self.db.get_record_by_id(concept_id, case_sensitive=False)
+        except DatabaseReadException as e:
+            logger.error(
+                f"Encountered DatabaseReadException looking up {concept_id}: {e}"
+            )
+        else:
+            if match:
+                self.add_record(response, match, match_type)
+            else:
+                logger.error(f"Unable to find expected record for {concept_id} matching as {match_type}")  # noqa: E501
 
     def post_process_resp(self, resp: Dict) -> Dict:
         """Fill all empty source_matches slots with NO_MATCH results and
         sort source records by descending `match_type`.
 
-        :param Dict resp: incoming response object
-        :return: response object with empty source slots filled with
-                NO_MATCH results and corresponding source metadata
+        :param resp: incoming response object
+        :return: response object with empty source slots filled with NO_MATCH results
+            and corresponding source metadata
         """
         for src_name in resp['source_matches'].keys():
             if resp['source_matches'][src_name] is None:
                 resp['source_matches'][src_name] = {
                     'match_type': MatchType.NO_MATCH,
                     'records': [],
-                    'source_meta_': self.fetch_meta(src_name)
+                    'source_meta_': self.db.get_source_metadata(src_name)
                 }
             else:
                 records = resp['source_matches'][src_name]['records']
@@ -215,8 +203,8 @@ class QueryHandler:
         """Return response as dict where key is source name and value
         is a list of records. Corresponds to `keyed=true` API parameter.
 
-        :param str query: string to match against
-        :param Set[str] sources: sources to match from
+        :param query: string to match against
+        :param sources: sources to match from
         :return: completed response object to return to client
         """
         resp = {
@@ -232,39 +220,35 @@ class QueryHandler:
 
         queries = list()
         if [p for p in PREFIX_LOOKUP.keys() if query_l.startswith(p)]:
-            pk = f'{query_l}##identity'
-            queries.append(pk)
+            queries.append((query_l, "identity"))
 
         for prefix in [p for p in NAMESPACE_LOOKUP.keys() if
                        query_l.startswith(p)]:
-            pk = f'{NAMESPACE_LOOKUP[prefix].lower()}:{query_l}##identity'
-            queries.append(pk)
+            term = f"{NAMESPACE_LOOKUP[prefix].lower()}:{query_l}"
+            queries.append((term, "identity"))
 
         for match in ITEM_TYPES.values():
-            pk = f'{query_l}##{match}'
-            queries.append(pk)
+            queries.append((query_l, match))
 
         matched_concept_ids = list()
-        for q in queries:
+        for term, item_type in queries:
             try:
-                query_resp = self.db.genes.query(
-                    KeyConditionExpression=Key('label_and_type').eq(q)
-                )
-                for record in query_resp['Items']:
-                    concept_id = record['concept_id']
-                    if concept_id in matched_concept_ids:
-                        continue
-                    else:
-                        if record['item_type'] == "identity":
-                            self.add_record(resp, record, MatchType.CONCEPT_ID)
-                        else:
-                            self.fetch_record(
-                                resp, concept_id,
-                                MatchType[record['item_type'].upper()])
-                        matched_concept_ids.append(concept_id)
+                if item_type == "identity":
+                    record = self.db.get_record_by_id(term, False)
+                    if record and record['concept_id'] not in matched_concept_ids:
+                        self.add_record(resp, record, MatchType.CONCEPT_ID)
+                else:
+                    refs = self.db.get_refs_by_type(term, RefType(item_type))
+                    for ref in refs:
+                        if ref not in matched_concept_ids:
+                            self.fetch_record(resp, ref, MatchType[item_type.upper()])
+                            matched_concept_ids.append(ref)
 
-            except ClientError as e:
-                logger.error(e.response['Error']['Message'])
+            except DatabaseReadException as e:
+                logger.error(
+                    f"Encountered DatabaseReadException looking up {item_type}"
+                    f" {term}: {e}"
+                )
                 continue
 
         # remaining sources get no match
@@ -274,8 +258,8 @@ class QueryHandler:
         """Return response as list, where the first key-value in each item
         is the source name. Corresponds to `keyed=false` API parameter.
 
-        :param str query: string to match against
-        :param List[str] sources: sources to match from
+        :param query: string to match against
+        :param sources: sources to match from
         :return: completed response object to return to client
         """
         response_dict = self.response_keyed(query, sources)
@@ -307,24 +291,22 @@ class QueryHandler:
                incl: str = '', excl: str = '', **params) -> SearchService:
         """Return highest match for each source.
 
-        :param str query_str: query, a string, to search for
-        :param bool keyed: if true, return response as dict keying source names
-            to source objects; otherwise, return list of source objects
-        :param str incl: str containing comma-separated names of sources to
-            use. Will exclude all other sources. Case-insensitive. Raises
-            InvalidParameterException if both incl and excl args are
-            provided, or if invalid source names are given.
-        :param str excl: str containing comma-separated names of source to
-            exclude. Will include all other source. Case-insensitive. Raises
-            InvalidParameterException if both incl and excl args are
-            provided, or if invalid source names are given.
+        :param query_str: query, a string, to search for
+        :param keyed: if true, return response as dict keying source names to source
+            objects; otherwise, return list of source objects
+        :param incl: str containing comma-separated names of sources to use. Will
+            exclude all other sources. Case-insensitive.
+        :param excl: str containing comma-separated names of source to exclude. Will
+            include all other source. Case-insensitive.
         :return: SearchService class containing all matches found in sources.
+        :raise InvalidParameterException: if both `incl` and `excl` args are provided,
+            or if invalid source names are given
         """
         possible_sources = {name.value.lower(): name.value for name in
                             SourceName.__members__.values()}
         sources = dict()
         for k, v in possible_sources.items():
-            if self.db.metadata.get_item(Key={'src_name': v}).get('Item'):
+            if self.db.get_source_metadata(v):
                 sources[k] = v
 
         if not incl and not excl:
@@ -372,7 +354,7 @@ class QueryHandler:
     def _add_merged_meta(self, response: NormalizeService) -> NormalizeService:
         """Add source metadata to response object.
 
-        :param Dict response: in-progress response object
+        :param response: in-progress response object
         :return: completed response object.
         """
         sources_meta = {}
@@ -383,7 +365,7 @@ class QueryHandler:
             prefix = concept_id.split(':')[0]
             src_name = PREFIX_LOOKUP[prefix.lower()]
             if src_name not in sources_meta:
-                sources_meta[src_name] = self.fetch_meta(src_name)
+                sources_meta[src_name] = self.db.get_source_metadata(src_name)
         response.source_meta_ = sources_meta
         return response
 
@@ -391,9 +373,9 @@ class QueryHandler:
                          possible_concepts: List[str]) -> NormService:
         """Add alternate matches warning to response object
 
-        :param NormService response: in-progress response object
-        :param Dict record: normalized record
-        :param List[str] possible_concepts: other possible matches
+        :param response: in-progress response object
+        :param record: normalized record
+        :param possible_concepts: other possible matches
         :return: updated response object
         """
         norm_concepts = set()
@@ -416,11 +398,10 @@ class QueryHandler:
     ) -> NormalizeService:
         """Add gene descriptor to response.
 
-        :param Dict response: Response object
-        :param Dict record: Gene record
-        :param MatchType match_type: query's match type
-        :param Optional[List[str]] possible_concepts: List of other normalized
-            concepts found
+        :param response: Response object
+        :param record: Gene record
+        :param match_type: query's match type
+        :param possible_concepts: List of other normalized concepts found
         :return: Response with gene descriptor
         """
         params = {
@@ -507,10 +488,10 @@ class QueryHandler:
         return response
 
     @staticmethod
-    def _record_order(record: Dict) -> (int, str):
+    def _record_order(record: Dict) -> Tuple[int, str]:
         """Construct priority order for matching. Only called by sort().
 
-        :param Dict record: individual record item in iterable to sort
+        :param record: individual record item in iterable to sort
         :return: tuple with rank value and concept ID
         """
         src = record['src_name'].upper()
@@ -518,12 +499,12 @@ class QueryHandler:
         return source_rank, record['concept_id']
 
     @staticmethod
-    def _handle_failed_merge_ref(record, response, query) -> Dict:
+    def _handle_failed_merge_ref(record: Dict, response: Dict, query: str) -> Dict:
         """Log + fill out response for a failed merge reference lookup.
 
-        :param Dict record: record containing failed merge_ref
-        :param Dict response: in-progress response object
-        :param str query: original query value
+        :param record: record containing failed merge_ref
+        :param response: in-progress response object
+        :param query: original query value
         :return: response with no match
         """
         logger.error(f"Merge ref lookup failed for ref {record['merge_ref']} "
@@ -534,7 +515,7 @@ class QueryHandler:
     def _prepare_normalized_response(self, query: str) -> Dict[str, Any]:
         """Provide base response object for normalize endpoints.
 
-        :param str query: user-provided query
+        :param query: user-provided query
         :return: basic normalization response boilerplate
         """
         return {
@@ -549,7 +530,7 @@ class QueryHandler:
     def normalize(self, query: str) -> NormalizeService:
         """Return normalized concept for query.
 
-        :param str query: String to find normalized concept for
+        :param query: String to find normalized concept for
         :return: Normalized gene concept
         """
         response = NormalizeService(**self._prepare_normalized_response(query))
@@ -562,11 +543,11 @@ class QueryHandler:
     ) -> NormService:
         """Given a record, return the corresponding normalized record
 
-        :param NormalizationService response: in-progress response object
-        :param Dict record: record to retrieve normalized concept for
-        :param MatchType match_type: type of match that returned these records
-        :param Callable callback: response constructor method
-        :param Optional[List[str]] possible_concepts: alternate possible matches
+        :param response: in-progress response object
+        :param record: record to retrieve normalized concept for
+        :param match_type: type of match that returned these records
+        :param callback: response constructor method
+        :param possible_concepts: alternate possible matches
         :return: Normalized response object
         """
         merge_ref = record.get("merge_ref")
@@ -586,24 +567,14 @@ class QueryHandler:
             # record is sole member of concept group
             return callback(response, record, match_type, possible_concepts)
 
-    def _get_matches_by_type(self, query: str, match_type: str) -> List[Dict]:
-        """Get matches list for match tier.
-        :param str query: user-provided query
-        :param str match_type: keyword of match type to check
-        :return: List of records matching the query and match level
-        """
-        matching_refs = self.db.get_records_by_type(query, match_type)
-        matching_records = [self.db.get_record_by_id(m["concept_id"], False)
-                            for m in matching_refs]
-        return sorted(matching_records, key=self._record_order)  # type: ignore
-
     def _perform_normalized_lookup(
         self, response: NormService, query: str, response_builder: Callable
     ) -> NormService:
         """Retrieve normalized concept, for use in normalization endpoints
-        :param NormService response: in-progress response object
-        :param str query: user-provided query
-        :param Callable response_builder: response constructor callback method
+
+        :param response: in-progress response object
+        :param query: user-provided query
+        :param response_builder: response constructor callback method
         :return: completed service response object
         """
         if query == "":
@@ -621,17 +592,15 @@ class QueryHandler:
             return self._resolve_merge(response, record, MatchType.CONCEPT_ID,
                                        response_builder)
 
-        for match_type in ITEM_TYPES.values():
+        for match_type in RefType:
             # get matches list for match tier
-            matching_refs = self.db.get_records_by_type(query_str, match_type)
+            matching_refs = self.db.get_refs_by_type(query_str, match_type)
             matching_records = \
-                [self.db.get_record_by_id(m['concept_id'], False)
-                 for m in matching_refs]
+                [self.db.get_record_by_id(ref, False) for ref in matching_refs]
             matching_records.sort(key=self._record_order)  # type: ignore
 
             if len(matching_refs) > 1:
-                possible_concepts = \
-                    [ref["concept_id"] for ref in matching_refs]
+                possible_concepts = [ref for ref in matching_refs]
             else:
                 possible_concepts = None
 
@@ -640,7 +609,7 @@ class QueryHandler:
                 assert match is not None
                 record = self.db.get_record_by_id(match["concept_id"], False)
                 if record:
-                    match_type_value = MatchType[match_type.upper()]
+                    match_type_value = MatchType[match_type.value.upper()]
                     return self._resolve_merge(
                         response, record, match_type_value,
                         response_builder, possible_concepts
@@ -653,20 +622,21 @@ class QueryHandler:
     ) -> UnmergedNormalizationService:
         """Add individual records to unmerged normalize response.
 
-        :param UnmergedNormalizationService response: in-progress response
-        :param Dict normalized_record: record associated with normalized concept,
-            either merged or single identity
-        :param MatchType match_type: type of match achieved
-        :param Optional[List[str]] possible_concepts: other possible results
+        :param response: in-progress response
+        :param normalized_record: record associated with normalized concept, either
+        merged or single identity
+        :param match_type: type of match achieved
+        :param possible_concepts: other possible results
         :return: Completed response object
         """
         response.match_type = match_type
         response.normalized_concept_id = normalized_record["concept_id"]
         if normalized_record["item_type"] == "identity":
             record_source = SourceName[normalized_record["src_name"].upper()]
+            meta = self.db.get_source_metadata(record_source.value)
             response.source_matches[record_source] = MatchesNormalized(
                 records=[BaseGene(**self._transform_locations(normalized_record))],
-                source_meta_=self.fetch_meta(record_source.value)
+                source_meta_=meta  # type: ignore
             )
         else:
             concept_ids = [normalized_record["concept_id"]] + \
@@ -680,9 +650,10 @@ class QueryHandler:
                 if record_source in response.source_matches:
                     response.source_matches[record_source].records.append(gene)
                 else:
+                    meta = self.db.get_source_metadata(record_source.value)
                     response.source_matches[record_source] = MatchesNormalized(
                         records=[gene],
-                        source_meta_=self.fetch_meta(record_source.value)
+                        source_meta_=meta,  # type: ignore
                     )
         if possible_concepts:
             response = self._add_alt_matches(response, normalized_record,
@@ -693,7 +664,7 @@ class QueryHandler:
         """Return all source records under the normalized concept for the
         provided query string.
 
-        :param str query: string to search against
+        :param query: string to search against
         :return: Normalized response object
         """
         response = UnmergedNormalizationService(
